from compress_fasttext import compress, decomposition, evaluation, navec_like, prune, quantization, utils
from compress_fasttext import models
from compress_fasttext.compress import make_new_fasttext_model, quantize_ft, svd_ft, prune_ft, prune_ft_freq
